// Copyright 2017 Archos SA
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//      http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.


package com.archos.filecorelibrary;

import java.io.Closeable;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.net.DatagramSocket;
import java.net.ServerSocket;

public final class IOUtils {

    /** close things if not null, for use in finally blocks, swallows exceptions */
    public static void closeSilently(final Closeable closeable) {
        if (closeable != null) {
            try {
                closeable.close();
            } catch (IOException e) {
                // silence
            }
        }
    }

    /** close things if not null, for use in finally blocks, swallows exceptions */
    public static void closeSilently(final ServerSocket socket) {
        if (socket != null && !socket.isClosed()) {
            try {
                socket.close();
            } catch (IOException e) {
                // silence
            }
        }
    }

    /** close things if not null, for use in finally blocks, swallows exceptions */
    public static void closeSilently(final DatagramSocket socket) {
        if (socket != null && !socket.isClosed()) {
            // does not throw
            socket.close();
        }
    }

    /**
     * Basic copy from one to another stream. Does not close streams or anything other than copying.
     * Creates an 8k buffer internally.
     */
    public static void streamCopy(final InputStream source, final OutputStream target) throws IOException {
        byte[] buffer = new byte[8 * 1024];
        int read;
        while ((read = source.read(buffer)) != -1) {
            target.write(buffer, 0, read);
        }
    }

    private IOUtils() {
        // empty
    }

}
