/*
 * Copyright 2017 Archos SA
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#define LOG_TAG "avos_media_player"

#include <stdio.h>
#include <string.h>
#include <dlfcn.h>
#include <pthread.h>
#include <sys/queue.h>
#include <stdlib.h>

#include <android/native_window_jni.h>

#include "jni.h"

int jniThrowException(C_JNIEnv* env, const char* className, const char* msg);
int jniGetFDFromFileDescriptor(C_JNIEnv* env, jobject fileDescriptor);

#include "libavos.h"

typedef struct event_msg {
    int what;
    int ext1;
    int ext2;
    avos_msg_t *extdata;
    TAILQ_ENTRY(event_msg) next;
} event_msg_t;
typedef TAILQ_HEAD(, event_msg) EVENT_MSG_QUEUE;

typedef struct event_ctx {
    jobject postevent_ref;

    pthread_mutex_t mtx;
    pthread_cond_t cond;
    pthread_t thread;

    EVENT_MSG_QUEUE msg_queue;
    int quit;
} event_ctx_t;

extern JavaVM *myVm;
static const avos_mp_handle_t *avos = NULL;
static const avos_metadata_handle_t *avos_metadata = NULL;

static struct {
    jfieldID    handle;
    jfieldID    native_window;
} mp_fields;

static pthread_mutex_t mtx = PTHREAD_MUTEX_INITIALIZER;

static inline int handle_ret(JNIEnv *env, int ret, const char *cmd)
{
    if (ret == AVOS_ERR_OK) {
        return ret;
    } else {
        char err_msg[256];

        snprintf(err_msg, 256, "%s returned %s", cmd,
          ret == AVOS_ERR ? "AVOS_ERR" : "AVOS_ERR_CRITICAL");
        jniThrowException(env, "java/lang/IllegalStateException", err_msg);
        return ret;
    }
}

#define CHECK(cmd) handle_ret(env, (cmd), #cmd)

static inline void set_mp(JNIEnv *env, jobject thiz, avos_mp_t *mp)
{
    pthread_mutex_lock(&mtx);
    (*env)->SetLongField(env, thiz, mp_fields.handle, (jlong) mp);
    pthread_mutex_unlock(&mtx);
}

static inline avos_mp_t *get_mp(JNIEnv *env, jobject thiz)
{
    avos_mp_t *mp;
    pthread_mutex_lock(&mtx);
    mp = (avos_mp_t *)(*env)->GetLongField(env, thiz, mp_fields.handle);
    pthread_mutex_unlock(&mtx);
    return mp;
}

static inline avos_mp_t *get_mp_or_throw(JNIEnv *env, jobject thiz)
{
    avos_mp_t *mp = get_mp(env, thiz);
    if (!mp)
        jniThrowException(env, "java/lang/IllegalStateException", "no mp object");
    return mp;
}

static inline void set_surface(JNIEnv *env, jobject thiz, void *surface)
{
    pthread_mutex_lock(&mtx);
    (*env)->SetLongField(env, thiz, mp_fields.native_window, (jlong)surface);
    pthread_mutex_unlock(&mtx);
}

static inline void *get_surface(JNIEnv *env, jobject thiz)
{
    void *surface;
    pthread_mutex_lock(&mtx);
    surface = (void *)(*env)->GetLongField(env, thiz, mp_fields.native_window);
    pthread_mutex_unlock(&mtx);
    return surface;
}

static void *event_thread(void *ctx)
{
    event_ctx_t *event_ctx = (event_ctx_t *) ctx;
    JNIEnv *env = NULL;
    event_msg_t *msg = NULL;
    jobject jextdata = NULL;

    LOGV("event_thread: start\n");

    (*myVm)->AttachCurrentThread(myVm, &env, NULL);

    while (!event_ctx->quit) {
        pthread_mutex_lock(&event_ctx->mtx);

        while (!event_ctx->quit && !(msg = TAILQ_FIRST(&event_ctx->msg_queue)))
            pthread_cond_wait(&event_ctx->cond, &event_ctx->mtx);

        if (event_ctx->quit) {
            // clean queue
            event_msg_t *msg_next;

            for (msg = TAILQ_FIRST(&event_ctx->msg_queue); msg != NULL; msg = msg_next) {
                msg_next = TAILQ_NEXT(msg, next);
                TAILQ_REMOVE(&event_ctx->msg_queue, msg, next);
                if (msg->extdata)
                    free(msg->extdata);
                free(msg);
            }
        } else if (msg) {
            TAILQ_REMOVE(&event_ctx->msg_queue, msg, next);
        }

        pthread_mutex_unlock(&event_ctx->mtx);

        if (!msg)
            continue;

        if (msg->extdata) {
            if (msg->extdata->type == AVOS_MSG_TYPE_STR) {
                jextdata = (jobject) (*env)->NewStringUTF(env, (const char *)msg->extdata->data);
            } else if (msg->extdata->type == AVOS_MSG_TYPE_TEXT_SUBTITLE) {
                avos_text_subtitle_t *sub = (avos_text_subtitle_t *)msg->extdata->data;
                jstring jstr = (jstring) (*env)->NewStringUTF(env, sub->text);
                if (jstr) {
                    jextdata = (*env)->CallStaticObjectMethod(env, fields.SubtitleClazz,
                            fields.Subtitle_createTimedTextSubtitleMethod,
                            sub->position, sub->duration, jstr);
                    (*env)->DeleteLocalRef(env, jstr);
                }
            } else if (msg->extdata->type == AVOS_MSG_TYPE_BITMAP_SUBTITLE) {
                avos_bitmap_subtitle_t *sub = (avos_bitmap_subtitle_t *)msg->extdata->data;
                jobject bitmap = create_bitmap(env, &sub->bitmap, 0, 0);
                if (bitmap) {
                    jextdata = (*env)->CallStaticObjectMethod(env, fields.SubtitleClazz,
                            fields.Subtitle_createTimedBitmapSubtitleMethod,
                            sub->position, sub->duration, sub->left_corner, sub->top_corner, sub->orig_width, sub->orig_height, bitmap);
                    (*env)->DeleteLocalRef(env, bitmap);
                }
            }
        }
        (*env)->CallStaticVoidMethod(env, fields.AvosMediaPlayerClazz,
                fields.AvosMediaPlayer_postEventMethod,
                event_ctx->postevent_ref,
                msg->what, msg->ext1, msg->ext2, jextdata);
        if (jextdata) {
             (*env)->DeleteLocalRef(env, jextdata);
             jextdata = NULL;
        }

        if (msg->extdata)
            free(msg->extdata);
        free(msg);
    }

    (*myVm)->DetachCurrentThread(myVm);
    LOGV("event_thread: end\n");
    return NULL;
}

static event_ctx_t *event_thread_create(JNIEnv *env, jobject postevent_ref)
{
    event_ctx_t *event_ctx = (event_ctx_t *)calloc(1, sizeof(event_ctx_t));

    if (!event_ctx)
        return NULL;

    event_ctx->postevent_ref = (*env)->NewGlobalRef(env, postevent_ref);
    pthread_mutex_init(&event_ctx->mtx, NULL);
    pthread_cond_init(&event_ctx->cond, NULL);
    TAILQ_INIT(&event_ctx->msg_queue);

    pthread_create(&event_ctx->thread, NULL, event_thread, event_ctx);
    return event_ctx;
}

static void event_thread_destroy(JNIEnv *env, event_ctx_t *event_ctx)
{
    pthread_mutex_lock(&event_ctx->mtx);
    event_ctx->quit = 1;
    pthread_cond_signal(&event_ctx->cond);
    pthread_mutex_unlock(&event_ctx->mtx);

    pthread_join(event_ctx->thread, NULL);

    LOGV("event_thread joined\n");

    pthread_mutex_destroy(&event_ctx->mtx);
    pthread_cond_destroy(&event_ctx->cond);
    (*env)->DeleteGlobalRef(env, event_ctx->postevent_ref);
    free(event_ctx);
}

static int event_thread_add(event_ctx_t *event_ctx, int what, int ext1, int ext2, avos_msg_t *extdata)
{
    event_msg_t *msg = (event_msg_t *)calloc(1, sizeof(event_msg_t));
    if (!msg)
        return -1;
    msg->what = what;
    msg->ext1 = ext1;
    msg->ext2 = ext2;
    msg->extdata = extdata;
    pthread_mutex_lock(&event_ctx->mtx);
    TAILQ_INSERT_TAIL(&event_ctx->msg_queue, msg, next);
    pthread_cond_signal(&event_ctx->cond);
    pthread_mutex_unlock(&event_ctx->mtx);
    return 0;
}

static void avos_mp_event_cb(avos_mp_t *mp, int what, int ext1, int ext2, avos_msg_t *extdata)
{
    int cleanup = 1;
    event_ctx_t *event_ctx = (event_ctx_t *)avos->getpriv(mp);

    if (event_ctx && event_thread_add(event_ctx, what, ext1, ext2, extdata) == 0)
        cleanup = 0;
    if (cleanup && extdata)
        free(extdata);
}

int
register_avosmediaplayer(JNIEnv *env)
{
    jclass clazz;

    clazz = (*env)->FindClass(env, "com/archos/medialib/AvosMediaPlayer");
    if (!clazz)
        return -1;
    mp_fields.handle = (*env)->GetFieldID(env, clazz, "mMediaPlayerHandle", "J");
    if (!mp_fields.handle)
        return -1;
    mp_fields.native_window = (*env)->GetFieldID(env, clazz, "mNativeWindowHandle", "J");
    if (!mp_fields.native_window)
        return -1;
    avos = avos_mp_get_handle();
    avos_metadata = avos_metadata_get_handle();
    return 0;
}

int
unregister_avosmediaplayer(JNIEnv *env)
{
    return 0;
}

void
Java_com_archos_medialib_AvosMediaPlayer_create(JNIEnv *env, jobject thiz, jobject weak_thiz)
{
    void *surface;
    const char *err_msg = NULL;
    avos_mp_t *mp;
    event_ctx_t *event_ctx;

    if (!avos)
        goto err;

    mp = avos->create(avos_mp_event_cb);
    if (!mp) {
        err_msg = "can't create mp";
        goto err;
    }

    event_ctx = event_thread_create(env, weak_thiz);
    if (!event_ctx) {
        err_msg = "can't create thread event";
        goto err;
    }
    avos->setpriv(mp, event_ctx);

    surface = get_surface(env, thiz);
    if (surface)
        avos->setsurface(mp, surface);

    set_mp(env, thiz, mp);
    return;
err:
    jniThrowException(env, "java/lang/IllegalStateException", err_msg);
}

static void
free_native_window(JNIEnv *env, jobject thiz)
{
    void *surface = get_surface(env, thiz);

    LOGV("free_native_window\n");

    if (surface) {
        ANativeWindow_release((ANativeWindow *)surface);
        set_surface(env, thiz, NULL);
    }
}

void
Java_com_archos_medialib_AvosMediaPlayer_setVideoSurface(JNIEnv *env, jobject thiz, jobject jsurface)
{
    int ret;
    void *surface;
    const char *err_msg = NULL;
    avos_mp_t *mp;
    jclass clazz;
    jfieldID surface_field;

    LOGD("setVideoSurface\n");

    free_native_window(env, thiz);

    surface = ANativeWindow_fromSurface(env, jsurface);;
    if (!surface) {
        err_msg = "!surface";
        goto err;
    }

    set_surface(env, thiz, surface);

    mp = get_mp(env, thiz);
    if (mp)
        avos->setsurface(mp, surface);
    return;
err:
    jniThrowException(env, "java/lang/IllegalArgumentException", err_msg);
}

void
Java_com_archos_medialib_AvosMediaPlayer_nativeRelease(JNIEnv *env, jobject thiz)
{
    event_ctx_t *event_ctx;
    avos_mp_t *mp = get_mp(env, thiz);
    if (!mp)
        return;
    LOGV("nativeRelease\n");

    event_ctx = (event_ctx_t *)avos->getpriv(mp);
    avos->destroy(mp);

    if (event_ctx)
        event_thread_destroy(env, event_ctx);

    free_native_window(env, thiz);

    set_mp(env, thiz, NULL);
}

void
Java_com_archos_medialib_AvosMediaPlayer_nativeReset(JNIEnv *env, jobject thiz)
{
    avos_mp_t *mp;
    void *priv = NULL;
    LOGV("nativeReset\n");

    free_native_window(env, thiz);
    mp = get_mp(env, thiz);
    if (mp) {
        priv = avos->getpriv(mp);
        avos->destroy(mp);
    }
    mp = avos->create(avos_mp_event_cb);
    if (!mp) {
        jniThrowException(env, "java/lang/IllegalStateException", "can't create mp");
        return;
    }
    avos->setpriv(mp, priv);

    set_mp(env, thiz, mp);
}

static void keys_values_free(char **keys)
{
    if (keys) {
        char **p = keys;
        while (*p) {
            LOGD("free: %s", *p);
            free(*p);
            p++;
        }
        free(keys);
    }
}


static int keys_values_fill(JNIEnv *env, jobjectArray keys, jobjectArray values,
        char ***p_keys, char ***p_values)
{
    int i;
    int nb_pairs = 0;
    char **c_keys = NULL;
    char **c_values = NULL;
    const char *c_key = NULL;
    const char *c_value = NULL;
    jstring key = NULL, value = NULL;

    if (keys == NULL || values == NULL)
        return -1;

    nb_pairs = (*env)->GetArrayLength(env, keys);
	if (nb_pairs != (*env)->GetArrayLength(env, values))
        return -1;


    c_keys = calloc(nb_pairs + 1, sizeof(const char *)); // NULL terminated
    c_values = calloc(nb_pairs + 1, sizeof(const char *)); // NULL terminated
    if (!c_keys || !c_values)
        goto err;


    for (i = 0; i < nb_pairs; ++i) {
        key = (jstring) (*env)->GetObjectArrayElement(env, keys, i);
        value = (jstring) (*env)->GetObjectArrayElement(env, values, i);
        if (!key || !value)
            goto err;
        c_key = (*env)->GetStringUTFChars(env, key, NULL);
        c_value = (*env)->GetStringUTFChars(env, value, NULL);
        if (!c_key || !c_value)
            goto err;
        c_keys[i] = strdup(c_key);
        c_values[i] = strdup(c_value);

        (*env)->ReleaseStringUTFChars(env, key, c_key);
        (*env)->ReleaseStringUTFChars(env, value, c_value);
        (*env)->DeleteLocalRef(env, key);
        (*env)->DeleteLocalRef(env, value);
        c_key = c_value = NULL;
        key = value = NULL;
    }
    *p_keys = c_keys;
    *p_values = c_values;

    return 0;
err:
    if (c_keys)
        keys_values_free(c_keys);
    if (c_values)
        keys_values_free(c_values);
    if (c_key)
        (*env)->ReleaseStringUTFChars(env, key, c_key);
    if (c_value)
        (*env)->ReleaseStringUTFChars(env, value, c_value);
    if (key)
        (*env)->DeleteLocalRef(env, key);
    if (value)
        (*env)->DeleteLocalRef(env, value);
    return -1;
}

void
Java_com_archos_medialib_AvosMediaPlayer_setDataSource(JNIEnv *env, jobject thiz, jstring path, jobjectArray keys, jobjectArray values)
{
    avos_mp_t *mp = get_mp_or_throw(env, thiz);
    if (!mp)
        return;

    if (path == NULL) {
        jniThrowException(env, "java/lang/IllegalArgumentException", "path is NULL");
        return;
    }

    const char *c_path = (*env)->GetStringUTFChars(env, path, NULL);
    if (c_path == NULL) {
        return;
    }
    char **c_keys = NULL;
    char **c_values = NULL;
    keys_values_fill(env, keys, values, &c_keys, &c_values);

    CHECK(avos->setdatasource(mp, c_path, (const char **)c_keys, (const char **)c_values));

    if (c_keys || c_values) {
        keys_values_free(c_keys);
        keys_values_free(c_values);
    }
    (*env)->ReleaseStringUTFChars(env, path, c_path);
}

void
Java_com_archos_medialib_AvosMediaPlayer_setDataSourceFD(JNIEnv *env, jobject thiz, jobject fileDescriptor, jlong offset, jlong length)
{
    avos_mp_t *mp = get_mp_or_throw(env, thiz);
    if (!mp)
        return;

    if (fileDescriptor == NULL) {
        jniThrowException(env, "java/lang/IllegalArgumentException", "fileDescriptor is NULL");
        return;
    }
    int fd = jniGetFDFromFileDescriptor(env, fileDescriptor);
    CHECK(avos->setdatasource_fd(mp, fd, offset, length));
}

void
Java_com_archos_medialib_AvosMediaPlayer_prepareAsync(JNIEnv *env, jobject thiz)
{
    avos_mp_t *mp = get_mp_or_throw(env, thiz);
    if (!mp) return;
    CHECK(avos->open_async(mp));
}

void
Java_com_archos_medialib_AvosMediaPlayer_prepare(JNIEnv *env, jobject thiz)
{
    avos_mp_t *mp = get_mp_or_throw(env, thiz);
    if (!mp) return;
    CHECK(avos->open(mp));
}

void
Java_com_archos_medialib_AvosMediaPlayer_nativeStart(JNIEnv *env, jobject thiz)
{
    avos_mp_t *mp = get_mp_or_throw(env, thiz);
    if (!mp) return;
    CHECK(avos->start(mp));
}

void
Java_com_archos_medialib_AvosMediaPlayer_nativeStop(JNIEnv *env, jobject thiz)
{
    avos_mp_t *mp = get_mp_or_throw(env, thiz);
    if (!mp) return;
    CHECK(avos->close(mp));
}

void
Java_com_archos_medialib_AvosMediaPlayer_nativePause(JNIEnv *env, jobject thiz)
{
    avos_mp_t *mp = get_mp_or_throw(env, thiz);
    if (!mp) return;
    CHECK(avos->pause(mp));
}

jboolean
Java_com_archos_medialib_AvosMediaPlayer_isPlaying(JNIEnv *env, jobject thiz)
{
    int ret = 0;
    avos_mp_t *mp = get_mp_or_throw(env, thiz);
    if (!mp) return 0;
    CHECK(avos->isplaying(mp, &ret));
    return ret;
}

void
Java_com_archos_medialib_AvosMediaPlayer_seekTo(JNIEnv *env, jobject thiz, int msec)
{
    avos_mp_t *mp = get_mp_or_throw(env, thiz);
    if (!mp) return;
    CHECK(avos->seek_async(mp, msec));
}

void
Java_com_archos_medialib_AvosMediaPlayer_nativeSetStartTime(JNIEnv *env, jobject thiz, int msec)
{
    avos_mp_t *mp = get_mp_or_throw(env, thiz);
    if (!mp) return;
    CHECK(avos->setstarttime(mp, msec));
}

int
Java_com_archos_medialib_AvosMediaPlayer_getCurrentPosition(JNIEnv *env, jobject thiz)
{
    uint32_t ret = 0;
    avos_mp_t *mp = get_mp_or_throw(env, thiz);
    if (!mp) return 0;
    CHECK(avos->getpos(mp, &ret));
    return ret;
}

int
Java_com_archos_medialib_AvosMediaPlayer_getBufferPosition(JNIEnv *env, jobject thiz)
{
    uint32_t ret = 0;
    avos_mp_t *mp = get_mp_or_throw(env, thiz);
    if (!mp) return 0;
    CHECK(avos->getpos(mp, &ret));
    return ret;
}

int
Java_com_archos_medialib_AvosMediaPlayer_getRelativePosition(JNIEnv *env, jobject thiz)
{
    uint32_t ret = 0;
    avos_mp_t *mp = get_mp_or_throw(env, thiz);
    if (!mp) return 0;
    CHECK(avos->getpos(mp, &ret));
    return ret;
}

int
Java_com_archos_medialib_AvosMediaPlayer_getDuration(JNIEnv *env, jobject thiz)
{
    uint32_t ret = 0;
    avos_mp_t *mp = get_mp_or_throw(env, thiz);
    if (!mp) return 0;
    CHECK(avos->getduration(mp, &ret));
    return ret;
}

void
Java_com_archos_medialib_AvosMediaPlayer_setLooping(JNIEnv *env, jobject thiz, jboolean looping)
{
    avos_mp_t *mp = get_mp_or_throw(env, thiz);
    if (!mp) return;
    CHECK(avos->setlooping(mp, looping));
}

jboolean
Java_com_archos_medialib_AvosMediaPlayer_isLooping(JNIEnv *env, jobject thiz)
{
    int ret = 0;
    avos_mp_t *mp = get_mp_or_throw(env, thiz);
    if (!mp) return 0;
    CHECK(avos->islooping(mp, &ret));
    return ret;
}

void
Java_com_archos_medialib_AvosMediaPlayer_setVolume(JNIEnv *env, jobject thiz, float leftVolume, float rightVolume)
{
    // XXX
}

int
Java_com_archos_medialib_AvosMediaPlayer_getAudioSessionId(JNIEnv *env, jobject thiz)
{
    int ret = 0;
    avos_mp_t *mp = get_mp_or_throw(env, thiz);
    if (!mp) return 0;
    CHECK(avos->getaudiosessionid(mp, &ret));
    return ret;
}

jbyteArray
Java_com_archos_medialib_AvosMediaPlayer_getMetadata(JNIEnv *env, jobject thiz)
{
    metadata_buffer_t *buffer = NULL;
    avos_mp_t *mp = get_mp_or_throw(env, thiz);
    jbyteArray array = NULL;

    if (!mp) return NULL;

    if (avos->getmetadata(mp, &buffer))
	    goto end;

    array = (*env)->NewByteArray(env, avos_metadata->size(buffer));
    if (!array)
	    goto end;

    jbyte* bytes = (*env)->GetByteArrayElements(env, array, NULL);
    if (!bytes)
	    goto end;

    memcpy(bytes, avos_metadata->data(buffer), avos_metadata->size(buffer));

    (*env)->ReleaseByteArrayElements(env, array, bytes, 0);

end:
    if (buffer)
	    avos_metadata->destroy(&buffer);

    return array;
}

jboolean
Java_com_archos_medialib_AvosMediaPlayer_setAudioTrack(JNIEnv *env, jobject thiz, int track)
{
    int ret = 0;
    avos_mp_t *mp = get_mp_or_throw(env, thiz);
    if (!mp) return 0;
    CHECK(avos->setaudiotrack(mp, track, &ret));
    return ret;
}

void
Java_com_archos_medialib_AvosMediaPlayer_checkSubtitles(JNIEnv *env, jobject thiz, int track)
{
    avos_mp_t *mp = get_mp_or_throw(env, thiz);
    if (!mp) return;
    CHECK(avos->checksubtitles(mp));
}

jboolean
Java_com_archos_medialib_AvosMediaPlayer_setSubtitleTrack(JNIEnv *env, jobject thiz, int track)
{
    int ret = 0;
    avos_mp_t *mp = get_mp_or_throw(env, thiz);
    if (!mp) return 0;
    CHECK(avos->setsubtitletrack(mp, track, &ret));
    return ret;
}

void
Java_com_archos_medialib_AvosMediaPlayer_setSubtitleDelay(JNIEnv *env, jobject thiz, int delay)
{
    avos_mp_t *mp = get_mp_or_throw(env, thiz);
    if (!mp) return;
    CHECK(avos->setsubtitledelay(mp, delay));
}

void
Java_com_archos_medialib_AvosMediaPlayer_setSubtitleRatio(JNIEnv *env, jobject thiz, int n, int d)
{
    avos_mp_t *mp = get_mp_or_throw(env, thiz);
    if (!mp) return;
    CHECK(avos->setsubtitleratio(mp, n, d));
}

void
Java_com_archos_medialib_AvosMediaPlayer_setAudioFilter(JNIEnv *env, jobject thiz, int n, int night_on)
{
    avos_mp_t *mp = get_mp_or_throw(env, thiz);
    if (!mp) return;
    CHECK(avos->setaudiofilter(mp, n, night_on));
}

void
Java_com_archos_medialib_AvosMediaPlayer_setAvDelay(JNIEnv *env, jobject thiz, int delay)
{
    avos_mp_t *mp = get_mp_or_throw(env, thiz);
    if (!mp) return;
    CHECK(avos->setavdelay(mp, delay));
}

void
Java_com_archos_medialib_AvosMediaPlayer_setAvSpeed(JNIEnv *env, jobject thiz, float speed)
{
    avos_mp_t *mp = get_mp_or_throw(env, thiz);
    if (!mp) return;
    CHECK(avos->setavspeed(mp, speed));
}

void
Java_com_archos_medialib_AvosMediaPlayer_setNextTrack(JNIEnv *env, jobject thiz, jstring path)
{
    const char *cPath = path ? (*env)->GetStringUTFChars(env, path, NULL) : NULL;
    avos_mp_t *mp = get_mp_or_throw(env, thiz);

    if (mp) {
        CHECK(avos->setnextrack(mp, cPath));
    }
    (*env)->ReleaseStringUTFChars(env, path, cPath);
}
